<?php

/*
 * ==========================================================
 * EXCHANGE
 * ==========================================================
 *
 *  2022-2023 boxcoin.dev. All rights reserved.
 *
 */

define('BXC_EXCHANGE', '1.0.3');

function bxc_exchange_quotation($send_amount, $send_code, $get_code) {
    $send_code = strtolower($send_code);
    $get_code = strtolower($get_code);
    $is_send_crypto = bxc_crypto_is($send_code);
    $is_get_crypto = bxc_crypto_is($get_code);
    $is_send_stablecoin = in_array($send_code, ['usdc', 'usdt']);
    $is_get_stablecoin = in_array($get_code, ['usdc', 'usdt']);
    $send_min_amount_usd = floatval(bxc_settings_get('exchange-min', 1));
    $send_decimals = $is_send_crypto && !$is_send_stablecoin ? bxc_crypto_get_decimals($send_code) : 2;
    $send_min_amount = $send_code == 'usd' ? $send_min_amount_usd : $send_min_amount_usd * bxc_exchange_rates('usd', $send_code);
    if ($send_amount < $send_min_amount) {
        return ['error' => 'amount-less-than-min', 'send_min_amount' => bxc_crypto_get_value_with_decimals($send_min_amount, $send_decimals)];
    }
    $fee = $is_get_crypto ? bxc_crypto_get_network_fee($get_code, $send_code) : 0;
    $fee_processing = $send_amount * (bxc_settings_get('exchange-fee', 0) / 100);
    $amount_less_fees = $send_amount - $fee_processing - $fee;
    $get_amount_to_send = $send_amount - $fee_processing;
    $balance = $is_get_crypto ? bxc_blockchain($get_code, 'balance') : 9999999;
    $get_max_amount = $balance * 0.98;
    $get_amount = bxc_crypto_get_cryptocurrency_value($amount_less_fees, $get_code, $send_code);
    if (bxc_settings_get('exchange-notifications-low-balance') && (bxc_crypto_get_fiat_value($balance, $get_code, 'usd') < bxc_settings_get('exchange-notifications-low-balance-threshold')) && (bxc_settings_db('exchange-notification-low-balance') < (time() - 86400))) {
        $language = bxc_settings_get('language-admin');
        $get_code_uppercase = strtoupper($get_code);
        bxc_email_notification(str_replace('{R}', $get_code_uppercase, bxc_m('Low {R} exchange balance', $language)), str_replace('{R}', $get_code_uppercase, bxc_m('The {R} exchange balance is low. Current balance: ', $language) . $balance . ' ' . $get_code_uppercase . '.'));
        bxc_settings_db('exchange-notification-low-balance', time());
    }
    if (bxc_crypto_get_cryptocurrency_value($get_amount_to_send, $get_code, $send_code) > $get_max_amount) {
        $send_max_amount = $is_send_crypto ? bxc_crypto_get_value_with_decimals(bxc_crypto_get_cryptocurrency_value($get_max_amount, $send_code, $get_code), $send_decimals) : bxc_crypto_get_fiat_value($get_max_amount, $get_code, $send_code);
        if (bxc_settings_get('exchange-notifications-exceeds-balance') && bxc_settings_db('exchange-notifications-exceeds-balance') != $send_amount) {
            $language = bxc_settings_get('language-admin');
            $send_code_uppercase = strtoupper($send_code);
            bxc_email_notification(bxc_m('Amount exceeds balance', $language), str_replace(['{R}', '{R2}'], [$send_amount . ' ' . $send_code_uppercase, $send_max_amount . ' ' . $send_code_uppercase], bxc_m('The exchange of {R} could not be completed because it exceeds the available balance of {R2}.', $language)));
            bxc_settings_db('exchange-notifications-exceeds-balance', $send_amount);
        }
        return ['error' => 'amount-exceeds-max', 'send_max_amount' => $send_max_amount];
    }
    $response = [
        'send_amount' => bxc_decimal_number($send_amount),
        'send_amount_less_fees' => bxc_crypto_get_value_with_decimals(bxc_decimal_number($amount_less_fees), $send_decimals),
        'get_amount' => $is_get_stablecoin ? bxc_crypto_get_value_with_decimals($get_amount, 2) : bxc_decimal_number($get_amount),
        'get_amount_to_send' => bxc_decimal_number(bxc_crypto_get_cryptocurrency_value($get_amount_to_send, $get_code, $send_code)),
        'unit_rate' => bxc_crypto_get_value_with_decimals(bxc_decimal_number(1 / bxc_exchange_rates($send_code, $get_code)), $send_decimals),
        'fee' => $fee ? bxc_crypto_get_value_with_decimals(bxc_decimal_number($fee), $send_decimals) : '0',
        'processing_fee' => bxc_crypto_get_value_with_decimals(bxc_decimal_number($fee_processing), $send_decimals),
        'send_code' => $send_code,
        'get_code' => $get_code
    ];
    return $response;
}

function bxc_exchange_is_payment_completed($external_reference_base64) {
    if (!empty($external_reference_base64)) {
        $transaction = bxc_exchange_get_transaction($external_reference_base64);
        return $transaction && in_array(bxc_isset($transaction, 'status'), ['X', 'C']);
    }
    return false;
}

function bxc_exchange_finalize($external_reference_base64, $identity = false, $manual_payment = false, $user_payment_details = false) {
    $error = 'Empty external reference.';
    if (!empty($external_reference_base64)) {

        // Identity verification
        if (!bxc_verify_admin() && bxc_settings_get('exchange-identity-type') == 'complycube') {
            if (!$identity || !isset($identity['email']) || !isset($identity['id'])) {
                return [false, 'Identity corrupted or not provided.'];
            }
            $complycube = bxc_isset(json_decode(bxc_settings_db('complycube', false, '{}'), true), $identity['email']);
            if (!$complycube || $complycube[0] !== $identity['id']) {
                return [false, 'Identity does not match or not found.'];
            }
        }

        // Get transaction
        $transaction = bxc_exchange_get_transaction($external_reference_base64);

        // Add customer
        if (!bxc_verify_admin() && ($identity && empty($transaction['customer_id']) && !empty($identity['email']))) {
            $customer_id = bxc_customers_get($identity['email']);
            if (!$customer_id) {
                $ip_info = bxc_ip_info('country,countryCode');
                $identity['country'] = bxc_isset($ip_info, 'country', '');
                $identity['country_code'] = bxc_isset($ip_info, 'countryCode', '');
                $customer_id = ['id' => bxc_customers_add(bxc_isset($identity, 'first_name'), bxc_isset($identity, 'last_name'), bxc_isset($identity, 'email', 'NULL'), bxc_isset($identity, 'phone'), bxc_isset($identity, 'country'), bxc_isset($identity, 'country_code'), bxc_isset($identity, 'extra'))];
            }
            bxc_transactions_update($transaction['id'], ['customer_id' => $customer_id['id']], false);
        }

        // Manual payment from user
        if ($manual_payment && $transaction) {
            $language = bxc_settings_get('language-admin');
            if (bxc_settings_get('exchange-notifications-completed')) {
                bxc_email_notification(bxc_m('New exchange requested', $language) . ' | ' . bxc_m('Pending payment', $language), str_replace(['{R}', '{R2}', '{R3}'], [$transaction['amount_fiat'] . ' ' . strtoupper($transaction['currency']), $transaction['amount'] . ' ' . strtoupper($transaction['cryptocurrency']), $transaction['id']], bxc_m('The exchange of {R} for {R2} has been requested. Boxcoin transaction ID: {R3}.', $language) . ' ' . bxc_m('The payment is pending and needs to be manually submitted by the user.', $language)));
            }
            return [true];
        }
        if ($transaction && in_array(bxc_isset($transaction, 'status'), ['X', 'C'])) {
            $history = json_decode(bxc_settings_db('exchange_transfers', false, '[]'), true);
            if (!in_array($transaction['id'], $history)) {
                $external_reference = explode('|', base64_decode($external_reference_base64));
                if (is_array($external_reference) && count($external_reference) > 2) {
                    $sending_currency = $external_reference[1];
                    $response = false;
                    $quote = bxc_exchange_quotation(bxc_isset($transaction, 'amount_fiat', $transaction['amount']), in_array($transaction['cryptocurrency'], ['stripe', 'verifone', 'paypal']) ? $transaction['currency'] : $transaction['cryptocurrency'], $sending_currency);
                    if (bxc_isset($quote, 'error')) {
                        $error = $quote['error'];
                    } else {
                        $network = bxc_crypto_get_network($sending_currency);
                        $get_amount = $quote['get_amount'];
                        if (bxc_crypto_is($sending_currency)) {

                            // Crypto payment sending
                            switch ($network) {
                                case 'btc':
                                    $response = bxc_btc_transfer($get_amount, $external_reference[2]);
                                    break;
                                case 'eth':
                                    $response = bxc_eth_transfer($get_amount, $sending_currency, $external_reference[2]);
                                    break;
                            }
                        } else {

                            // FIAT payment sending
                            $response = 'manual';
                        }
                        if ($response && is_string($response) && !strpos($response, 'error')) {
                            array_push($history, $transaction['id']);
                            bxc_settings_db('exchange_transfers', $history);
                            $language = bxc_settings_get('language-admin');
                            $transaction_update_details = [];
                            $description = bxc_transactions_get_description($transaction['id']);
                            if (bxc_settings_get('exchange-notifications-completed')) {
                                bxc_email_notification(bxc_m($response == 'manual' ? 'New exchange requested' : 'New exchange completed', $language), str_replace(['{R}', '{R2}', '{R3}'], [$quote['send_amount'] . ' ' . strtoupper($quote['send_code']), $get_amount . ' ' . strtoupper($sending_currency), $transaction['id']], bxc_m($response == 'manual' ? 'The exchange of {R} for {R2} has been requested. Boxcoin transaction ID: {R3}.' : 'The exchange of {R} for {R2} has been completed. Boxcoin transaction ID: {R3}.', $language)));
                            }
                            if ($user_payment_details) {
                                array_push($description, bxc_m('Payment details of the user', $language) . ':' . PHP_EOL . bxc_isset($user_payment_details, 'details', '') . PHP_EOL . bxc_m('Payment method', $language) . ': ' . bxc_m(bxc_isset($user_payment_details, 'method_name', ''), $language));
                            }
                            if (($network == 'btc' && bxc_settings_get('testnet-btc')) || ($network == 'eth' && bxc_settings_get('testnet-eth'))) {
                                array_push($description, bxc_m('Testnet transaction', $language));
                            }
                            if (bxc_crypto_is_fiat($transaction['cryptocurrency'])) {
                                array_push($description, bxc_m('Payment method', $language) . ': ' . ucfirst($transaction['cryptocurrency']));
                            }
                            $transaction_update_details['to'] = $external_reference[2];
                            $transaction_update_details['cryptocurrency'] = $sending_currency;
                            $transaction_update_details['amount'] = $get_amount;
                            $transaction_update_details['hash'] = $response == 'manual' ? '' : $response;
                            $transaction_update_details['description'] = $description;
                            bxc_transactions_update($transaction['id'], $transaction_update_details, false);
                            return [true, $get_amount, $response];
                        } else {
                            $error = is_string($response) ? $response : json_encode($response);
                        }
                    }
                } else {
                    $error = bxc_error('Bad external reference: ' . base64_decode($external_reference_base64), 'bxc_exchange_finalize', true);
                }
            } else {
                return [true, $transaction['amount'], bxc_crypto_is($transaction['cryptocurrency']) ? $transaction['hash'] : 'manual', 'Exchange already executed.'];
            }
        } else {
            $error = 'Transaction not completed.';
        }
    }

    // Notifications
    if (bxc_settings_get('exchange-notifications-failed')) {
        $language = bxc_settings_get('language-admin');
        bxc_email_notification(bxc_m('Exchange failed', $language), str_replace(['{R}', '{R2}'], [$transaction['id'], $error], bxc_m('An exchange was unsuccessful. Boxcoin transaction ID: {R}. Error: {R2}.', $language)));
    }

    return [false, $error];
}

function bxc_exchange_finalize_manual($amount, $cryptocurrency_code, $currency_code, $external_reference, $note, $identity = false, $user_payment_details = false) {
    $response = bxc_transactions_create($amount, $cryptocurrency_code, $currency_code, $external_reference, '', $note, false, '', false, false, $identity, false, 3);
    return is_array($response) && is_numeric($response[0]) ? bxc_exchange_finalize($external_reference, $identity, true, $user_payment_details) : [false, $response];
}

function bxc_exchange_get_transaction($external_reference_base64) {
    return bxc_db_get('SELECT * FROM bxc_transactions WHERE external_reference = "' . bxc_db_escape($external_reference_base64) . '" ORDER BY id DESC LIMIT 1');
}

function bxc_complycube_curl($url_part, $body = [], $type = 'POST') {
    return json_decode(bxc_curl('https://api.complycube.com/v1/' . $url_part, json_encode($body), ['Authorization: ' . bxc_settings_get('exchange-complycube-api-key'), 'Content-Type: application/json'], $type), true);
}

function bxc_complycube($first_name, $last_name, $email) {
    $response = bxc_complycube_curl('clients', ['type' => 'person', 'email' => $email, 'personDetails' => ['firstName' => $first_name, 'lastName' => $last_name]]);
    $id = bxc_isset($response, 'id');
    if ($id) {
        $response = bxc_complycube_curl('tokens', ['clientId' => $id, 'referrer' => '*://*/*']);
        $token = bxc_isset($response, 'token');
        return $token ? [true, $token, $id] : [false, $response];
    }
    return [false, $response];
}

function bxc_complycube_create_check($client_id, $live_photo_id, $document_id) {
    $response = bxc_complycube_curl('checks', ['clientId' => $client_id, 'livePhotoId' => $live_photo_id, 'documentId' => $document_id, 'type' => 'identity_check']);
    return isset($response['status']) ? [true, $response] : [false, $response];
}

function bxc_complycube_check($check_id, $email) {
    $response = bxc_complycube_curl('checks/' . $check_id, [], 'GET');
    $status = bxc_isset($response, 'status');
    if ($status) {
        if ($status == 'complete') {
            $complycube = json_decode(bxc_settings_db('complycube', false, '{}'), true);
            $complycube[$email] = [$check_id, time()];
            bxc_settings_db('complycube', $complycube);
            return true;
        }
        return [true, $status];
    }
    return [false, $response];
}

function bxc_exchange_email_verification($email, $saved_email = false, $verification_code = false) {
    if ($saved_email) {
        $saved_email = bxc_encryption($saved_email, false);
        if ($saved_email) {
            $verified_emails = json_decode(bxc_settings_db('verified_emails', false, '{}'), true);
            if (in_array($saved_email, $verified_emails) && $saved_email == $email) {
                return true;
            }
        }
    }
    if ($verification_code && $verification_code[0] && $verification_code[1]) {
        $response = trim($verification_code[0]) === bxc_encryption($verification_code[1], false);
        if ($response) {
            $verified_emails = json_decode(bxc_settings_db('verified_emails', false, '{}'), true);
            array_push($verified_emails, bxc_encryption($verification_code[2], false));
            bxc_settings_db('verified_emails', $verified_emails);
        }
        return $response;
    } else {
        $code = rand(99999, 999999);
        $text = bxc_('Your email verification code is') . ' ' . $code;
        $response = bxc_email_send($email, $text, ['message' => $text], true);
        return $response === true ? [true, bxc_encryption($code), bxc_encryption($email)] : [false, $response];
    }
}

?>